from abc import ABC, abstractmethod
import inspect
from typing import Any
from django.db import models
from django.db import models

from ..bases import UTZDecorator
from .exceptions import ModelConfigurationError



class ModelDecorator(UTZDecorator, ABC):
    """
    Base class for all `django_utz` model decorators. 
    """
    all_configs = ()
    required_configs = ()
    __slots__ = ("model",)

    def __init__(self, model: type[models.Model]) -> None:
        self.model = self.check_model(model)
        super().__init__()
    

    def __call__(self) -> type[models.Model]:
        prepared_model = self.prepare_model()
        if not issubclass(prepared_model, models.Model):
            raise TypeError("prepare_model method must return a model")
        return prepared_model


    def check_model(self, model: type[models.Model]) -> type[models.Model]:
        """
        Check if model and model configuration is valid. Returns the model if it is valid.

        `self.model` should not be accessed in this method as it set by this method.

        :param model: The model to check
        :return: The model if it is valid
        """
        if not issubclass(model, models.Model):
            raise TypeError(f"{model.__name__} is not a Django model")
        
        if not hasattr(model, "UTZMeta"):
            raise AttributeError("Model must have a UTZMeta class")
        
        if not inspect.isclass(model.UTZMeta):
            raise ModelConfigurationError("UTZMeta must be a class")
        
        for config in self.required_configs:
            if not getattr(model.UTZMeta, config, None):
                raise ModelConfigurationError(f"'{config}' must be set in the model's UTZMeta class")
        return model


    @abstractmethod
    def prepare_model(self) -> type[models.Model]:
        """
        Prepare the model for use. This where you can customize the model.

        :return: The model
        """
        pass

    
    def get_config(self, attr: str, default: Any = None) -> Any | None:
        """
        Get the value of a configuration attribute from the model.

        :param attr: The name of the configuration attribute
        :param default: The default value to return if the attribute is not set
        :return: The value of the configuration attribute
        """
        if not hasattr(self, "model"):
            raise AttributeError("Model not set")
        
        val = getattr(self.model.UTZMeta, attr, default)
        if val is not None and hasattr(self, f"validate_{attr}"):
            getattr(self, f"validate_{attr}")(val)
        return val


    def set_config(self, attr: str, value: Any) -> None:
        """
        Set the value of a configuration attribute in the model.
        
        :param attr: The name of the configuration attribute
        :param value: The value to set
        """
        if not hasattr(self, "model"):
            raise AttributeError("Model not set")
        
        if attr != "_decorated" and attr not in self.all_configs:
            raise ModelConfigurationError(f"Invalid config: {attr}")
        
        if value is not None and hasattr(self, f"validate_{attr}"):
            getattr(self, f"validate_{attr}")(value)

        setattr(self.model.UTZMeta, attr, value)
        return None




from typing import Callable
from django.db import models
from django.conf import settings

from ..utils import get_attr_by_traversal
from ...middleware import get_request_user
from ...datetime import utzdatetime
from .exceptions import ModelError, ModelConfigurationError


def get_model_traversal_path(model: type[models.Model]) -> str:
    """Returns the traversal path to the model."""
    return f"{model.__module__}.{model.__name__}"


def get_project_user_model() -> str:
    """Returns the traversal path to the project's user model."""
    auth_user_model: str = settings.AUTH_USER_MODEL
    split = auth_user_model.split('.', maxsplit=1)
    split.insert(1, "models")
    user_model_path = ".".join(split)
    return user_model_path


def is_user_model(model: type[models.Model]) -> bool:
    """Check if the model is the project's user model."""
    return get_project_user_model() == get_model_traversal_path(model)



class FunctionAttribute:
    """
    A descriptor class that returns the result of 
    passing an object into a function as an attribute.
    """
    def __init__(self, func: Callable[..., utzdatetime]):
        if not callable(func):
            raise TypeError("Value must be a function.")
        self.func = func


    def __set_name__(self, objtype, name: str):
        if not isinstance(name, str):
            raise TypeError('name must be a string')
        self.name = name


    def __get__(self, obj, objtype):
        return self.func(obj)



def get_user(model_obj: models.Model) -> models.Model | None:
    """
    Gets and returns the user object whose timezone is to be used by the model instance.
    
    If  the `use_related_user_timezone` config is True and `related_user` is not set, the first
    user object found that is related to the model instance is returned.

    If the user has still not been found, the request user is returned.

    :param model_obj: The model instance for which a user object is to be found.
    :return: The user object.
    :raises ModelError: If the user's model is not decorated with a `ModelDecorator`.
    """
    # Initially, assume that the object passed is a user object.
    user = model_obj
    model_obj_cls = model_obj.__class__
    if not is_user_model(model_obj_cls):
        if model_obj_cls.UTZMeta.use_related_user_timezone:
            # Get the user object as specified by the `related_user` config 
            # or by finding the user field in the model and its related models
            related_user = getattr(model_obj_cls.UTZMeta, "related_user", None) or find_user_field(model_obj_cls)
            if not related_user:
                if getattr(model_obj_cls.UTZMeta, "related_user", None):
                    raise ModelConfigurationError(
                        f"Please set the `related_user` config to the name or traversal path of the user field in {model_obj_cls.__name__}."
                    )
                raise ModelError(f"No relation to the User model was found in {model_obj_cls.__name__}")
            
            user = get_attr_by_traversal(obj=model_obj, traversal_path=related_user)

        else:
            user = get_request_user()
            
    if user:
        user_model_is_decorated = getattr(user.__class__, "UTZMeta", None) and getattr(user.__class__.UTZMeta, "_decorated", False)
        if not user_model_is_decorated:
            raise ModelError(
                f"Model '{user.__class__.__name__}', has not been decorated with a `ModelDecorator`"
            )
    return user



def find_user_field(model: type[models.Model]) -> str | None:
    """
    Finds and returns the traversal path to the first user field found
    in the model or its related models.

    This method assumes that the user is a foreign key or one-to-one field.
    """
    field_paths = []

    def find_user_related_field(model: type[models.Model]) -> str | None:
        """
        Finds and returns the user related field traversal path 
        in the given model or its related models.
        """
        for field in model._meta.fields:
            related_model = field.related_model
            if related_model and isinstance(field, (models.ForeignKey, models.OneToOneField)):
                field_paths.append(field.name)

                if is_user_model(related_model):
                    return ".".join(field_paths)
                else:
                    # if the field is not the user model, recursively check the field's related_model for the user field
                    return find_user_related_field(related_model)  
        field_paths.pop()
        return None
            
    return find_user_related_field(model)



from abc import ABC, abstractmethod
from typing import Any


class UTZDecorator(ABC):
    """Abstract base class for all `django_utz` decorators."""
    all_configs = ()
    required_configs = ()
    __slots__ = ()

    def __init__(self) -> None:
        self.set_config("_decorated", True)

    @abstractmethod
    def __call__(self) -> Any:
        """
        The method that is called when the decorator is used on a model or serializer.

        :return: The decorated model or serializer.
        """
        pass

    @abstractmethod
    def set_config(self, attr: str, value: Any) -> None:
        """
        Sets a configuration attribute on the model or serializer.

        :param attr: The attribute to set.
        :param value: The value to set.
        """
        pass

    @abstractmethod
    def get_config(self, attr: str, default: Any = None) -> Any | None:
        """
        Get the value of a configuration attribute from the model or serializer.

        :param attr: The attribute to get.
        :param default: The default value to return if the attribute is not set.
        :return: The value of the attribute or `default`.
        """
        pass




"""
`django_utz` decorators for django rest framework model serializers and custom serializer fields.
"""
try:
    import rest_framework
except ImportError:
    raise Exception(
        "Using django_utz.decorators.serializers requires the installation of the django rest framework.\
         Perhaps you forgot to install it. Run `pip install djangorestframework` and\
         add 'rest_framework' to settings.INSTALLED_APPS to continue."
        )

from .decorators import ModelSerializerDecorator, modelserializer



"""Middleware that aims to make the current request user available globally in the application."""

from django.utils.deprecation import MiddlewareMixin
import threading
from django.contrib.auth.models import AbstractUser, User, AbstractBaseUser
from typing import TypeVar

ProjectUserModel = TypeVar("ProjectUserModel", AbstractUser, User, AbstractBaseUser)

local_thread_storage = threading.local()


class DjangoUTZMiddleware(MiddlewareMixin):
    
    def process_request(self, request) -> None:
        """Stores the current authenticated user in the local thread storage."""
        if request.user.is_authenticated:
            setattr(local_thread_storage, "django_utz-request_user", request.user)
        else:
            setattr(local_thread_storage, "django_utz-request_user", None)
        

    def process_response(self, request, response):
        """
        Deletes the current authenticated user from the local thread storage.
        """
        if hasattr(local_thread_storage, "django_utz-request_user"):
            delattr(local_thread_storage, "django_utz-request_user")
        return response


def get_request_user() -> ProjectUserModel | None:
    """Returns the currently authenticated user. If no user is authenticated, returns None."""
    if hasattr(local_thread_storage, "django_utz-request_user"):
        request_user = getattr(local_thread_storage, "django_utz-request_user")
        if request_user is not None and request_user.is_authenticated:
            return request_user     
    return None



"""
#### django_utz

Decorators for Django models and DRF model serializers, custom serializer fields, 
template tags and filters that aid easy conversion of timezone aware fields to a user's timezone.

@Author: Daniel T. Afolayan (ti-oluwa.github.io)
"""

__version__ = "0.1.6"
__author__ = "Daniel T. Afolayan"

alias = "django-user-timezone"

try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo
except:
    raise Exception(
        "django_utz couldn't import the zoneinfo module.\
         Perhaps you are on an older version of Python, run `pip install backports.zoneinfo` to continue."
        )




from ...base_exceptions import UTZError, ConfigurationError


class SerializerError(UTZError):
    """There was an error with the model"""
    pass


class SerializerConfigurationError(ConfigurationError):
    """There was an error with the model's configuration"""
    pass



from django.utils import timezone
import datetime



class utzdatetime(datetime.datetime):
    """Custom datetime class that can be independent of settings.USE_TZ"""
    convert_to_local_time = False

    @classmethod
    def from_datetime(cls, _datetime: datetime.datetime):
        """Constructs and returns a `utzdatetime` object from a `datetime.datetime` object"""
        if not isinstance(_datetime, datetime.datetime):
            raise TypeError(f"_datetime expected datetime.datetime object, got {type(_datetime)}")
        if timezone.is_naive(_datetime):
            default_timezone = timezone.get_default_timezone()
            _datetime = timezone.make_aware(_datetime, default_timezone)
        return cls(
            year=_datetime.year,
            month=_datetime.month,
            day=_datetime.day,
            hour=_datetime.hour,
            minute=_datetime.minute,
            second=_datetime.second,
            microsecond=_datetime.microsecond,
            tzinfo=_datetime.tzinfo
        )


    def regard_usetz(self):
        """Respect `settings.USE_TZ` when converting to local time"""
        self.convert_to_local_time = True
        return self


    def disregard_usetz(self):
        """Remain unaffected by `settings.USE_TZ` when converting to local time"""
        self.convert_to_local_time = False
        return self





"""Utility functions for `django_utz` decorators"""
try:
    import zoneinfo
except:
    from backports import zoneinfo
from typing import Any, Callable
import pytz
import datetime
from django.db import models
from django.core.exceptions import ValidationError
import functools

from .bases import UTZDecorator


def validate_timezone(value: Any) -> str | datetime.tzinfo:
    """
    Validator to check if a timezone is valid.

    :param value: Timezone value to validate
    :raises: ValidationError if the value is not a valid timezone
    :return: The value if it is a valid timezone
    """
    if not is_timezone_valid(value):
        raise ValidationError("Invalid timezone.")
    return value


def is_timezone_valid(timezone: str | datetime.tzinfo) -> bool:
    """
    Returns whether timezone is a valid timezone.

    :param timezone: timezone info or name
    :return: True if timezone is a valid, False otherwise.
    """
    if isinstance(timezone, datetime.tzinfo):
        return True
    is_pytz = False
    is_zoneinfo = False
    try:
        pytz.timezone(timezone)
        is_pytz = True
    except pytz.exceptions.UnknownTimeZoneError:
        pass
    try:
        zoneinfo.ZoneInfo(timezone)
        is_zoneinfo = True
    except Exception:
        pass
    return is_pytz or is_zoneinfo


def get_attr_by_traversal(obj: object, traversal_path: str, default=None) -> object | Any | None:
    """
    Get an attribute of an object by traversing the object using the traversal path.
    
    :param obj: The object to traverse.
    :param traversal_path: The traversal path to the attribute. For example: "b.c.d"
    :param default: The default value to return if the attribute is not found.

    #### For example:
    ```
    class A:
        b = B()
    
    class B:
        c = C()

    class C:
        d = D()

    class D:
        def __init__(self):
            self.x = 10
    
    traversal_path = "b.c.d.x"
    a = A()
    x = get_attr_by_traversal(a, traversal_path)
    print(x) # 10
    ```
    """
    try:
        attrs = traversal_path.split('.')
        for attr in attrs:
            obj = getattr(obj, attr)
        return obj
    except AttributeError:
        return default


def is_datetime_field(model: type[models.Model], field_name: str) -> bool:
    """
    Checks if the given field name is a datetime field in the model.

    :param model: The model to check.
    :param field_name: The name of the field to check.
    :return: True if the field is a datetime field, False otherwise.
    """
    field = model._meta.get_field(field_name)
    return issubclass(field.__class__, models.DateTimeField)


def is_time_field(model: type[models.Model], field_name: str) -> bool:
    """
    Checks if the given field name is a time field in the model.

    :param model: The model to check.
    :param field_name: The name of the field to check.
    :return: True if the field is a time field, False otherwise.
    """
    field = model._meta.get_field(field_name)
    return issubclass(field.__class__, models.TimeField)


def is_date_field(model: type[models.Model], field_name: str) -> bool:
    """
    Checks if the given field name is a date field in the model.

    :param model: The model to check.
    :param field_name: The name of the field to check.
    :return: True if the field is a date field, False otherwise.
    """
    field = model._meta.get_field(field_name)
    return issubclass(field.__class__, models.DateField)



def transform_utz_decorator(decorator: type[UTZDecorator]) -> Callable[[type[object]], type[object]]:
    """
    Transforms class type utz decorator to a function type decorator.

    Using a utz decorator directly on a class will require that the class be instantiated
    before the decorator logic is applied. This is not ideal as the decorator logic should be
    applied to the decorated class itself and a modified class should be returned.

    A decorator wrapper function is returned that takes a class, applies the wrapped utz decorator logic,
    and returns the modified class.
    """
    @functools.wraps(decorator)
    def decorator_wrapper(cls: type[object]) -> type[object]:
        """Wrapper function that applies the utz decorator to the decorated class."""
        return decorator(cls)()
    
    return decorator_wrapper



from django.db import models
import pytz
import datetime
from django.db import models
from django.conf import settings

try:
    import zoneinfo
except:
    from backports import zoneinfo
from django.contrib.auth.models import AbstractUser, AbstractBaseUser, User
from typing import TypeVar, List
from django.core.exceptions import FieldDoesNotExist
from typing import Callable


from ..utils import is_datetime_field, is_timezone_valid, validate_timezone, transform_utz_decorator
from ...datetime import utzdatetime
from .exceptions import ModelError, ModelConfigurationError
from .bases import ModelDecorator
from .utils import get_user, is_user_model, FunctionAttribute



class UserModelUTZMixin:
    """Adds neccessary methods and properties to the User Model"""
    
    @property
    def utz(self):
        """
        The user's timezone info as a `pytz.tzinfo` or `zoneinfo.ZoneInfo` object.

        if settings.USE_DEPRECATED_PYTZ is True, then the user's timezone info is returned as a pytz.tzinfo object.
        Otherwise, it is returned as a zoneinfo.ZoneInfo object.
        """
        utz = getattr(self, self.UTZMeta.timezone_field)
        
        if utz and isinstance(utz, datetime.tzinfo):
            zone_str = str(utz)

        elif utz and isinstance(utz, str):
            if not is_timezone_valid(utz):
                raise ValueError(f"Invalid timezone: {utz}.")
            zone_str = utz

        if getattr(settings, "USE_DEPRECATED_PYTZ", False):
            return pytz.timezone(zone_str)

        return zoneinfo.ZoneInfo(zone_str)


    def to_local_timezone(self, _datetime: datetime.datetime) -> utzdatetime:
        """
        Adjust datetime.datetime object to user's local timezone.
        If the user's timezone is not set, the datetime object is returned as is.
        
        :param _datetime: `datetime.datetime` object
        :return: `utzdatetime` object

        datetime object returned is not affected by `settings.USE_TZ`.
        To change this behavior do:
        ```
        import datetime
        utz_datetime = user.to_local_timezone(datetime.datetime.now())
        utz_datetime.regard_usetz()
        ```
        """
        utz_dt = utzdatetime.from_datetime(_datetime)
        user_tz = self.utz
        if user_tz:
            return utz_dt.astimezone(user_tz)
        return utz_dt


UserModel = TypeVar("UserModel", AbstractBaseUser, AbstractUser, User)
UTZUserModel = TypeVar("UTZUserModel", AbstractBaseUser, AbstractUser, User, UserModelUTZMixin)


class UserModelDecorator(ModelDecorator):
    """`ModelDecorator` for the project's user model."""
    all_configs = ("timezone_field",)
    required_configs = ("timezone_field",)

    def __init__(self, model: UserModel) -> None:
        super().__init__(model)


    def check_model(self, model: UserModel) -> UserModel:
        """Ensures that the model in which this mixin is used is the project's user model"""
        if not is_user_model(model):
            raise ModelError(f"Model '{model.__name__}' is not the project's user model")  
        return super().check_model(model)
    

    def validate_timezone_field(self, value: str) -> None:
        if not isinstance(value, str):
            raise ModelConfigurationError("Value for 'timezone_field' should be of type str")
        return None


    def prepare_model(self) -> UTZUserModel:
        # Add timezone validator to the `timezone_field`` if not already added
        try:
            timezone_field = self.get_config("timezone_field")
            field = self.model._meta.get_field(timezone_field)
            if field and validate_timezone not in field.validators:
                field.validators = [*field.validators, validate_timezone]
        except FieldDoesNotExist:
            raise ModelError(f"Field '{timezone_field}' does not exist in model '{self.model.__name__}'")
        
        self.model.__bases__ = (UserModelUTZMixin, *self.model.__bases__)
        return self.model



class RegularModelDecorator(ModelDecorator):
    """`ModelDecorator` for regular Django models."""
    all_configs = ("datetime_fields", "attribute_suffix", "use_related_user_timezone", "related_user")
    required_configs = ("datetime_fields",)

    def check_model(self, model: models.Model) -> models.Model:
        model = super().check_model(model)

        related_user = getattr(model.UTZMeta, "related_user", None)
        if related_user and not isinstance(related_user, str):
            raise ModelConfigurationError("'related_user' should be of type str")
        return model
    

    def prepare_model(self) -> type[models.Model]:
        if self.get_config("datetime_fields") == "__all__":
            self.set_config("datetime_fields", self.get_datetime_fields(self.model))

        if not self.get_config("attribute_suffix"):
            self.set_config("attribute_suffix", "utz")

        if not self.get_config("use_related_user_timezone"):
            self.set_config("use_related_user_timezone", False)

        return self.update_model_attrs(self.model)
    

    def validate_datetime_fields(self, value: str | List[str] | tuple[str]) -> None:
        if value != "__all__" and not isinstance(value, (list, tuple)):
            raise ModelConfigurationError(
                f"'datetime_fields' should be a list, tuple or '__all__'"
            )
        return None
    

    def validate_attribute_suffix(self, value: str) -> None:
        if not isinstance(value, str):
            raise ModelConfigurationError("'attribute_suffix' should be of type str")
        return None
    

    def validate_use_related_user_timezone(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise ModelConfigurationError("'use_related_user_timezone' should be of type bool")
        return None
    

    def get_datetime_fields(self, model: type[models.Model]) -> List[str]:
        """Returns the datetime fields in the given model."""
        return [field.name for field in model._meta.fields if isinstance(field, models.DateTimeField)]
    

    def make_func_for_field(self, datetime_field: str) -> Callable[[models.Model], utzdatetime]:
        """
        Makes and returns a function that returns the value of the datetime field on a model instance
        in the user's local timezone.

        :param datetime_field: The name of the datetime field for which to make the function
        """
        func_name = f"get_{datetime_field}_{self.get_config("attribute_suffix")}"

        def func(model_instance: models.Model) -> utzdatetime:
            user: UserModel = get_user(model_instance)
            if user is None:
                return utzdatetime.from_datetime(getattr(model_instance, datetime_field))
            return user.to_local_timezone(getattr(model_instance, datetime_field))
        
        func.__name__ = func_name
        func.__qualname__ = func_name
        return func
            

    def update_model_attrs(self, model: type[models.Model]) -> type[models.Model]:
        """
        Updates the model with the read-only attributes for the datetime fields.

        :param model: The model to update
        """
        datetime_fields: List[str] = self.get_config("datetime_fields")
        attribute_suffix: str = self.get_config("attribute_suffix")

        for field in datetime_fields:
            if not is_datetime_field(model, field):
                raise ModelConfigurationError(f"Field '{field}' is not a datetime field in model '{model.__name__}'")
            
            func = self.make_func_for_field(field)
            read_only_attr = FunctionAttribute(func)
            attr_name = f"{field}_{attribute_suffix}"
            setattr(model, attr_name, read_only_attr)
        return model



# Function-type decorator for django models

def model(model: type[models.Model]) -> type[models.Model]:
    """
    #### `django_utz` decorator for django models.

    The decorated model must have a `UTZMeta` class with the following attributes:
    - `datetime_fields`: A list of the names of the datetime fields in the model or "__all__" to use all datetime fields.
    - `attribute_suffix`: Optional. The suffix to be added to the read-only attributes for utz versions  the datetime fields.
    Defaults to "utz".
    - `use_related_user_timezone`: Optional. A boolean indicating whether to use the timezone of the related user model.
    Defaults to False.
    - `related_user`: Optional. The name of the related user model if `use_related_user_timezone` is True.

    Example Usage:
    ```
    from django.db import models
    from django_utz.decorators import model

    @model
    class Article(models.Model):
        title = models.CharField(max_length=100)
        content = models.TextField()
        author = models.ForeignKey("Author", on_delete=models.CASCADE)
        published_at = models.DateTimeField()
        updated_at = models.DateTimeField()

        class UTZMeta:
            datetime_fields = ("published_at", "updated_at")
            attribute_suffix = "local"
    ```
    """
    return transform_utz_decorator(RegularModelDecorator)(model)



def usermodel(model: UserModel) -> UTZUserModel:
    """
    #### `django_utz` decorator for the project's user model.

    The user model must be decorated with this decorator for `django_utz` to work properly.
    The decorated model must have a `UTZMeta` class with the following attributes:
    - `timezone_field`: The name of the timezone field in the model.

    Example Usage:
    ```
    from django.contrib.auth.models import AbstractUser
    from django_utz.decorators import usermodel

    @usermodel
    class User(AbstractUser):
        id = models.AutoField(primary_key=True)
        email = models.EmailField(unique=True)
        username = models.CharField(max_length=100)
        ...
        timezone = models.CharField(max_length=100)

        class UTZMeta:
            timezone_field = "timezone"
    ```
    """
    return transform_utz_decorator(UserModelDecorator)(model)



from django.contrib.auth import get_user_model
from django.dispatch import Signal, receiver
from django.db.models.signals import pre_save



ProjectUserModel = get_user_model()

user_timezone_changed = Signal()
# ----------------------------------------------------
# This signal is sent when a user's timezone is changed.
# kwargs: instance, previous_timezone, current_timezone
# ----------------------------------------------------

@receiver(pre_save, sender=ProjectUserModel)
def utz_change_handler(sender, **kwargs) -> None:
    """Signal handler for user timezone changes"""
    user = kwargs["instance"]
    utz_meta = getattr(sender, "UTZMeta", None)

    if utz_meta and utz_meta._decorated:
        tz_field = getattr(utz_meta, "timezone_field")
        try:
            previous_timezone = getattr(sender.objects.get(pk=user.pk), tz_field)
        except sender.DoesNotExist:
            previous_timezone = ''
        current_timezone = getattr(user, tz_field)
        
        if previous_timezone != current_timezone:
            user_timezone_changed.send(
                sender=sender, 
                instance=user, 
                previous_timezone=previous_timezone, 
                current_timezone=current_timezone
            )
    return None



from rest_framework import serializers
from django.db import models
from typing import Any, Dict
import inspect
import datetime

from ..bases import UTZDecorator
from ..utils import transform_utz_decorator
from ..models.exceptions import ModelError
from .exceptions import SerializerConfigurationError
from ...serializer_fields import UTZDateTimeField



class ModelSerializerDecorator(UTZDecorator):
    """Decorator for `rest_framework.serializers.ModelSerializer` classes."""
    all_configs = ("auto_add_fields", "datetime_format")
    required_configs = ("auto_add_fields",)
    __slots__ = ("serializer",)

    def __init__(self, serializer: type[serializers.ModelSerializer]) -> None:
        self.serializer = self.check_model_serializer(serializer)
        super().__init__()


    @property
    def serializer_model(self) -> type[models.Model]:
        """
        Returns the serializer's model.
        """
        return self.serializer.Meta.model
    
    
    def __call__(self) -> type[models.Model]:
        prepared_serializer = self.prepare_serializer()
        if not issubclass(prepared_serializer, serializers.ModelSerializer):
            raise TypeError("prepare_serializer method must return a model serializer")
        return prepared_serializer
    

    def check_model_serializer(self, serializer: type[serializers.ModelSerializer]) -> type[serializers.ModelSerializer]:
        """
        Check if the model serializer is properly setup.

        :param serializer: The model serializer to check.
        :return: The model serializer if properly setup.
        """
        if not issubclass(serializer, serializers.ModelSerializer):
            raise TypeError(f"'{serializer.__name__}' is not a model serializer")
        
        model = serializer.Meta.model
        utz_meta = getattr(model, "UTZMeta", None)
        model_is_decorated = utz_meta and utz_meta._decorated and getattr(utz_meta, "datetime_fields", None)
        if not model_is_decorated:
            raise ModelError(
                f"Serializer's model {model.__name__}, has not been decorated with a `ModelDecorator`"
            )
        
        if not hasattr(serializer, "UTZMeta"):
            raise AttributeError("Model serializer must have a UTZMeta class")
        
        if not inspect.isclass(model.UTZMeta):
            raise SerializerConfigurationError("UTZMeta must be a class")
        
        for config in self.required_configs:
            if not getattr(serializer.UTZMeta, config, None):
                raise SerializerConfigurationError(f"'{config}' must be set in the model serializer's UTZMeta class")
        return serializer
    

    def validate_auto_add_fields(self, value: Any) -> None:
        if not isinstance(value, bool):
            raise SerializerConfigurationError("auto_add_fields must be a boolean")
        return None


    def validate_datetime_format(self, value: Any) -> None:
        if not isinstance(value, str):
            raise SerializerConfigurationError("datetime_format must be a string")

        try:
            datetime.datetime.now().strftime(value)
        except ValueError:
            raise SerializerConfigurationError(f"Invalid datetime format: {value}")
        return None


    def prepare_serializer(self) -> type[serializers.ModelSerializer]:
        """
        Prepare the serializer for use. This where you can customize the serializer.

        Here, user timezone aware datetime fields are automatically added to the serializer class.
        If `auto_add_fields` is set as True.
        """
        datetime_fields = self.serializer_model.UTZMeta.datetime_fields
        serializer_meta = self.serializer.Meta
        excluded_fields = serializer_meta.exclude if hasattr(serializer_meta, "exclude") else []
        included_fields = serializer_meta.fields if hasattr(serializer_meta, "fields") else []

        if included_fields == "__all__":
            included_fields = self.serializer_model._meta.fields

        # If the serializer has no fields, return it as is
        if not included_fields:
            return self.serializer

        if self.get_config("auto_add_fields"):
            for fieldname in datetime_fields:
                # Skip fields that are not included in the serializer
                if fieldname not in included_fields or fieldname in excluded_fields:
                    continue
                # If field is already defined in the serializer, skip it
                if hasattr(self.serializer, fieldname):
                    continue
                utz_datetime_field = self.get_utzdatetimefield_for_field(fieldname)
                setattr(self.serializer, fieldname, utz_datetime_field)
        return self.serializer
    

    def get_utzdatetimefield_for_field(self, datetime_field: str, serializer_meta: type[object]) -> UTZDateTimeField:
        """
        Returns a `UTZDateTimeField` for the given datetime field name in a serializer.

        :param datetime_field: The name of the datetime model field for which a `UTZDateTimeField` will be returned. 
        :param serializer_meta: The model serializer's meta class.
        :return: a `django_utz.serializer_fields.UTZDateTimeField` for the given datetime field.
        """
        meta_extra_kwargs: Dict = serializer_meta.extra_kwargs if hasattr(serializer_meta, "extra_kwargs") else {}
        field_extra_kwargs: Dict = meta_extra_kwargs.get(datetime_field, {}) # Get extra kwargs for the field if any

        datetime_format = self.get_config("datetime_format")
        if "format" not in field_extra_kwargs and datetime_format:
            field_extra_kwargs.update({"format": datetime_format}) 
        return UTZDateTimeField(**field_extra_kwargs)


    def set_config(self, attr: str, value: Any) -> None:
        """
        Sets a configuration attribute on the serializer.

        :param attr: The attribute to set.
        :param value: The value to set.
        """
        if not hasattr(self, "serializer"):
            raise AttributeError("Serializer not set")
        
        if attr != "_decorated" and attr not in self.all_configs:
            raise SerializerConfigurationError(f"Invalid config: {attr}")
        
        if value is not None and hasattr(self, f"validate_{attr}"):
            getattr(self, f"validate_{attr}")(value)

        setattr(self.serializer.UTZMeta, attr, value)
        return None


    def get_config(self, attr: str, default: Any = None) -> Any | None:
        """
        Gets a configuration attribute from the serializer.

        :param attr: The attribute to get.
        :param default: The default value to return if the attribute is not set.
        :return: The value of the attribute.
        """
        if not hasattr(self, "serializer"):
            raise AttributeError("Serializer not set")
        
        val = getattr(self.serializer.UTZMeta, attr, default)
        if val is not None and hasattr(self, f"validate_{attr}"):
            getattr(self, f"validate_{attr}")(val)
        return val


# Funtion-type decorator for `rest_framework.serializers.ModelSerializer` classes

def modelserializer(serializer: type[serializers.ModelSerializer]) -> type[serializers.ModelSerializer]:
    """
    #### `django_utz` decorator for `reest_framework.serializers.ModelSerializer` classes.

    This decorator auto adds user timezone aware datetime fields to the serializer class 
    in-place of the normal datetime fields, if `auto_add_fields` config is set as True.

    The decorated model serializer must have a `UTZMeta` class with the following attributes:
    - `auto_add_fields`: Optional. A boolean that determines if user timezone aware datetime fields should be automatically added to the serializer.
    - `datetime_format`: Optional. A string that determines the output format for the user timezone aware datetime fields.

    Example Usage:
    ```python
    from rest_framework import serializers
    from django_utz.decorators import modelserializer

    from .models import Article

    @modelserializer
    class ArticleSerializer(serializers.ModelSerializer):
        class Meta:
            model = Article
            fields = "__all__"
            exclude = ("updated_at",)
            extra_kwargs = {
                "published_at": {"read_only": True},
            }

        class UTZMeta:
            auto_add_fields = True
            datetime_format = "%Y-%m-%d %H:%M:%S"
    ```
    """
    return transform_utz_decorator(ModelSerializerDecorator)(serializer)



from ...base_exceptions import UTZError, ConfigurationError


class ModelError(UTZError):
    """There was an error with the model"""
    pass


class ModelConfigurationError(ConfigurationError):
    """There was an error with the model's configuration"""
    pass



"""django_utz template tags and filters"""

from typing import Dict, List
from django.template.base import Parser, Token, Node, FilterExpression, kwarg_re
from django.template import Library, TemplateSyntaxError, NodeList
from django.conf import settings
from django.utils import timezone
import datetime

from ..middleware import get_request_user
from ..decorators.models import UserModelUTZMixin

register = Library()
_generic_name = "usertimezone"


def parse_tag(token: Token, parser: Parser):
    """
    Generic template tag parser.

    Returns a three-tuple: (tag_name, args, kwargs)

    tag_name is a string, the name of the tag.

    args is a list of FilterExpressions, from all the arguments that didn't look like kwargs,
    in the order they occurred, including any that were amongst kwargs.

    kwargs is a dictionary mapping kwarg names to FilterExpressions, for all the arguments that
    looked like kwargs, including any that were mingled amongst args.

    (At rendering time, a FilterExpression f can be evaluated by calling f.resolve(context).)
    """
    bits = token.split_contents()
    tag_name = bits.pop(0)

    # Parse the rest of the args, and build FilterExpressions from them so that
    # we can evaluate them later.
    args = []
    kwargs = {}
    for bit in bits:
        # Is this a kwarg or an arg?
        match = kwarg_re.match(bit)
        kwarg_format = match and match.group(1)
        if kwarg_format:
            key, value = match.groups()
            kwargs[key] = FilterExpression(value, parser)
        else:
            args.append(FilterExpression(bit, parser))
    return tag_name, args, kwargs



class UTZNode(Node):
    """
    Node to render the datetimes in template content 
    in the preferred user's timezone.
    """
    def __init__(
            self, 
            nodelist: NodeList, 
            args: List[FilterExpression], 
            kwargs: Dict[str, FilterExpression], 
            tag_name: str
        ) -> None:
        self.nodelist = nodelist
        self.user = self.get_user(args, kwargs, tag_name)
        self.tag_name = tag_name


    @staticmethod
    def get_user(
        args: List[FilterExpression], 
        kwargs: Dict[str, FilterExpression], 
        tag_name: str
    ) -> FilterExpression | None:
        """
        Get the user object from the arguments and 
        keyword arguments passed to the utz template tag.
        """
        if args:
            if len(args) != 1:
                raise TemplateSyntaxError(
                    f"{tag_name} requires exactly one argument. The user object\
                    whose timezone is to be used in rendering the template content."
                )
            user = args[0]
        else:
            user = kwargs.get("user", None)
        return user


    def render(self, context):
        request_user = context.get("request").user
        if self.user:
            preferred_user = self.user.resolve(context)
        user = preferred_user if preferred_user else request_user

        utz_meta = getattr(user, "UTZMeta", None)
        user_model_is_decorated = utz_meta and utz_meta._decorated and issubclass(user.__class__, UserModelUTZMixin)
        if user.is_authenticated and user_model_is_decorated:
            tz = user.utz
            original_timezone = timezone.get_current_timezone()
            try:
                timezone.activate(tz)
                output = self.nodelist.render(context)
            finally:
                timezone.activate(original_timezone)
        else:
            output = self.nodelist.render(context)

        return output



@register.tag(name=_generic_name)
def utz_tag(parser: Parser, token: Token):
    """
    Template tag to render datetimes in the template content 
    in the preferred user's timezone.

    #### In your template:
    To render datetimes in the template content in the timezone 
    of the user object passed as an argument:
    ```
    {% load utz %}
    {% usertimezone user=object.user %}
        {# Your template content #}
    {% endusertimezone %}
    ```

    To render the template content in the timezone of the request user:
    ```
    {% load utz %}
    {% usertimezone %}
        {# Your template content #}
    {% endusertimezone %}
    ```
    """
    tag_name, args, kwargs = parse_tag(token, parser)
    nodelist = parser.parse(parse_until=(f"end{_generic_name}",))
    parser.delete_first_token()
    return UTZNode(nodelist, args, kwargs, tag_name)



@register.filter(name=_generic_name)
def utz_filter(value: datetime.datetime, user: UserModelUTZMixin = None):
    """
    Filter to convert a datetime object to the request/provided user's timezone.

    If the user is not authenticated or the user model is not a subclass of UTZUserModelMixin,
    the value is returned as is.

    #### In your template:
    ```
    {% load utz %}
    {{ datetime_object|usertimezone }}
    ```
    A specific user whose timezone would be used, can also be provided:
    ```
    {% load utz %}
    {{ datetime_object|usertimezone:user }}
    ```
    """
    if not user: # If no user is provided, use the request user
        user = get_request_user()
    
    utz_meta = getattr(user, "UTZMeta", None)
    is_decorated = utz_meta and utz_meta._decorated and issubclass(user.__class__, UserModelUTZMixin)
    # If the user model was not decorated with the `usermodel` decorator, return the value as is
    if not is_decorated:
        return value
    try:
        return user.to_local_timezone(value)
    except:
        return value
    



"""Provides custom serializer fields for converting datetime to user's local timezone on representation"""
import zoneinfo
from rest_framework import serializers
from django.conf import settings
import pytz
from django.db import models

from .decorators.models.utils import get_user
from .decorators.models.exceptions import ModelError


class UTZField:
    """
    Mixin for converting native serializer fields to a UTZField.
    """
    def __new__(cls, *args, **kwargs):
        new_field = super().__new__(*args, **kwargs)
        new_field._call_count = 0
        return new_field
    

    def to_representation(self, value):
        instance = self.root.instance
        # If the instance is a queryset, get the current instance from the queryset
        if issubclass(instance.__class__, (models.QuerySet, list)):
            instance = instance[self._call_count]
            self._call_count += 1
            
        instance_model_is_decorated = getattr(instance.__class__, "UTZMeta", None) and getattr(instance.__class__.UTZMeta, "_decorated", False)
        if not instance_model_is_decorated:
            raise ModelError(
                f"Model '{instance.__class__.__name__}', has not been decorated with a `ModelDecorator`"
            )
        
        user = get_user(instance)
        if user:
            self.timezone = user.utz
        return super().to_representation(value)
    

    def to_internal_value(self, value):
        if getattr(settings, "USE_DEPRACATED_PYTZ", False):
            server_timezone = pytz.timezone(settings.TIME_ZONE) if settings.USE_TZ else pytz.utc
        else:
            server_timezone = zoneinfo.ZoneInfo(settings.TIME_ZONE) if settings.USE_TZ else zoneinfo.ZoneInfo("UTC")
        return super().to_internal_value(value).astimezone(server_timezone)



class UTZDateTimeField(UTZField, serializers.DateTimeField):
    """
    Custom `serializers.DateTimeField` that converts input datetime to server's timezone(settings.TIMEZONE) before storage
    and converts back to the preferred user's local timezone on output.

    However, the model of the serializer in which this field is used must be decorated with a `ModelDecorator`.
    
    This field is usually added automatically to the model serializer if it is deorated with a `ModelSerializerDecorator`. 
    This can also work as a standalone field in a serializer without the decorator.

    if settings.USE_TZ is False, server's timezone is assumed to be UTC

    #### Example:
    ```
    class BookSerializer(serializers.ModelSerializer):
        created_at = UTZDateTimeField(format="%Y-%m-%d %H:%M:%S %Z (%z)")
        updated_at = UTZDateTimeField(format="%Y-%m-%d %H:%M:%S %Z (%z)")

        class Meta:
            model = Book
            fields = "__all__"
            extra_kwargs = {
                "created_at": {"read_only": True},
                "updated_at": {"read_only": True},
            }
    ```
    """
    pass
